// *********************************************************************
// *                                                                   *
// * (c) 1997-2004 Andreas Ziermann LOET.DE                            *
// *                                                                   *
// * These files are for reference only. You may not use these files   *
// * for commercial use without permit.                                *
// *                                                                   *
// * b@loet.de                                                         *
// *                                                                   *
// * 1.00 2004-07-04                                                   *
// *      first released                                               *
// *                                                                   *
// *********************************************************************

#include <stddef.h>
#include <malloc.h>
#include <dos.h>
#include <mem.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <conio.h>

#define FLAG_CF 1
typedef unsigned char byte;

// #define VOIDPTR void *
#define VOIDPTR void far *

bool print_alloc=false;

bool heap_free_count()
{
  REGPACK r;

  r.r_ax=0x4800;
  r.r_bx=0xffff;
  intr(0x21,&r); // largest block

  char buf[80];
  if (r.r_ax==7)
  {
    // heap corrupt
    sprintf(buf,"%dos_heap fehlerhaft: %d\n",r.r_ax);
    printf(buf);
    asm int 3;
    return false;
  }
  ::gotoxy(2,2);
  sprintf(buf,"groesster freier block: 0x%04x  %d KByte\n",r.r_bx,r.r_bx/64);
  printf(buf);
  return true;
}

#ifdef DEBUG
enum ALLOC_TYPE
{
  ALLOC_BY_MALLOC,
  ALLOC_BY_NEW,
  ALLOC_BY_NEW_KLAMMER,
};

VOIDPTR heap_malloc(size_t size,ALLOC_TYPE type);

VOIDPTR operator new(size_t size)
{
  return heap_malloc(size,ALLOC_BY_NEW);
}

VOIDPTR operator new[](size_t size)
{
  return heap_malloc(size,ALLOC_BY_NEW_KLAMMER);
}

VOIDPTR malloc(size_t size)
{
  return heap_malloc(size,ALLOC_BY_MALLOC);
}

void operator delete (VOIDPTR ptr)
{
  if (ptr)
    free(ptr);
}

void operator delete [](VOIDPTR ptr)
{
  if (ptr)
    free(ptr);
}

struct InfoBlock
{
  InfoBlock far *prev;
  unsigned long size;
  ALLOC_TYPE    type;
};

InfoBlock far *prev=NULL;

byte pre[16]={0xbe,0xbe,0xbe,0xbe,0xbe,0xbe,0xbe,0xbe,
              0xbe,0xbe,0xbe,0xbe,0xbe,0xbe,0xbe,0xbe};
byte post[16]={0xb0,0xb0,0xb0,0xb0,0xb0,0xb0,0xb0,0xb0,
              0xb0,0xb0,0xb0,0xb0,0xb0,0xb0,0xb0,0xb0};

bool HeapCheck()
{
  REGPACK r;

  r.r_ax=0x4800;
  r.r_bx=0xffff;
  intr(0x21,&r); // largest block

  if (r.r_ax==7)
  {
    // heap corrupt
    asm int 3;
    return false;
  }
  return true;
}

bool heap_check()
{
  if (!HeapCheck())
  {
    return false;
  } // ist der dos-heap noch ok?

  if (!prev)
    return true;

  InfoBlock far *info_header=prev;

  while (info_header)
  {
    byte far(*pre_bytes)[16]=(byte far(*)[16])info_header;
    pre_bytes=(byte far (*)[16])(((byte far *)pre_bytes)+16);
    if (memcmp(pre_bytes,pre,sizeof(pre))!=0)
    {
      asm int 3;
      return false;
    }
    byte far(*post_bytes)[16]=(byte far (*)[16])info_header;
    post_bytes=(byte far(*)[16])(((byte far *)post_bytes)+32+info_header->size);
    if (memcmp(post_bytes,post,sizeof(post))!=0)
    {
      asm int 3;
      return false;
    }
    info_header=info_header->prev;
  }
  return true;
}

void free(VOIDPTR ptr)
{
  if (!ptr)
  {
    // achtung delete(ptr) ruft einfach free(ptr) auf
    // ist aber das argument NULL eigentlich nicht definiert
    // zumindest nicht portabel
    asm int 3;
    return;
  }
  heap_check();

  bool found=false;
  InfoBlock far *info_header=prev;

  while (info_header)
  {
    void far *myptr=(byte *)info_header+32;
    if (myptr==ptr)
    {
      found=true;
      break;
    }
    info_header=info_header->prev;
  };
  if (!found)
  {
    asm int 3;
    return;
  }
  if (print_alloc==true)
  {
    char buf[256];
    sprintf(buf,"free     0x%04x%04x:0x%08x\n",FP_SEG(ptr),FP_OFF(ptr),info_header->size);
    print_alloc=false;
    printf(buf);
    print_alloc=true;
  }
  info_header=(InfoBlock far *)((byte *)ptr-32);
  memset(ptr,0xf0,info_header->size);
  // asm int 3;
}

VOIDPTR heap_malloc(size_t size,ALLOC_TYPE type)
{

  if (!heap_check())
    return NULL;

  REGPACK r;

  r.r_ax=0x4800;
  r.r_bx=0xffff;
  intr(0x21,&r); // largest block

  if (size>65000l)
  {
    asm int 3;
     // to heavy
  }

  if (size!=0)
    size--;
  r.r_ax=0x4800;
  r.r_bx=(size>>4)+4; // 3 *16 byte mehr anfordern
  intr(0x21,&r);
  if (r.r_flags & FLAG_CF)
  {
    asm int 3;
    // kann nicht allokieren
    return 0;
  }
  size++;
  InfoBlock far *info_header=(InfoBlock far *)MK_FP(r.r_ax,0);
  _fmemset(info_header,0xcc,16);
  info_header->prev=prev;
  info_header->size=size;
  info_header->type=type;
  prev=info_header;
  void far *pre_bytes=MK_FP(r.r_ax,16);
  _fmemset(pre_bytes,0xbe,16);
  void far *post_bytes=MK_FP(r.r_ax,size+32);
  _fmemset(post_bytes,0xb0,16);
  void far *ptr=MK_FP(r.r_ax,32);
  _fmemset(ptr,0xfe,size);

  if (print_alloc==true)
  {
    char buf[256];
    sprintf(buf,"alloc    0x%04x%04x:0x%08x\n",FP_SEG(ptr),FP_OFF(ptr),size);
    print_alloc=false;
    printf(buf);
    print_alloc=true;
  }

  return (VOIDPTR)ptr;
}

#endif

void gotoxy(int x,int y)
{
  if (x<1 || x>80)
    asm int 3;
  if (y<1 || y>25)
    asm int 3;
  x--;
  y--;

  x&=0xff;
  y&=0xff;

  REGPACK r;
  r.r_ax=0x200;
  r.r_bx=0x0;
  r.r_dx=(y<<8)|x;
  intr(0x10,&r);
}

int sprintf(char *buf,const char *fmt,...)
{
  va_list argptr;
  int cnt=0;

  va_start(argptr, fmt);

  const char *fmt_ptr=fmt;
  char c=*fmt_ptr;
  char *dest=buf;
  *dest=0;

  while (c)
  {
    c=*fmt_ptr++;
    if (c=='%')
    {
      c=*(fmt_ptr++);
      if (c=='%')
      {
        *dest++=c;
      }
      else
      {
        int count=0; // count parameter
        bool leading_zero=false;
      myswitch:
        switch(c)
        {
          case 's':
          {
            char *arg = va_arg(argptr,char *);
            int len=strlen(arg);
            if (count)
            {
              memset(dest,' ',count);
              if (len>count)
              {
                arg+=(len-count); // rechtsbuendig & abschneiden
                len=count;
              }
              dest+=(count-len);
            }
            strcpy(dest,arg);
            dest+=len;
            *dest=0;
            break;
          }
          case 'd':
          case 'x':
          {
            int base=10;
            if (c=='x')
              base=16;
            int arg = va_arg(argptr,int);
            int len=1;
            int _arg=arg;
            while (_arg>=base)
            {
              len++;
              _arg/=base;
            }

            if (count)
            {
              memset(dest,leading_zero?'0':' ',count);
              // if (len>count) dann wird einfach abgebrochen
              // dazu spaeter mehr
              len=count;
            }
            int mycount=len;
            char *_dest=dest+len;
            *_dest--=0;
            _arg=arg;
            do
            {
              int rest=_arg%base;
              if (rest<=9)
                *_dest--='0'+rest;
              else
                *_dest--='A'+rest-10;
              _arg/=base;
              mycount--;
            } while ( arg>=base && mycount );
            dest+=len;

            break;
          }
          default:
          {
            if (c=='0')
              leading_zero=true;
            while (c>='0' && c<='9')
            {
              count*=10;
              count+=c-'0';
              c=*fmt_ptr++;
            }
            if (!count)
              asm int 3; // unknown(unhandled) %switch
          }
          if (count) goto myswitch;
        }
        c=*fmt_ptr;
      }
    }
    else
    {
      // c!='%'
      *dest++=c;
      *dest=0;
    }
  };
  va_end(argptr);
  return(cnt);
}

size_t strlen(const char *src)
{
  size_t ret=0;   
  while(*src++)
    ret++;
  return ret;
}

