// *********************************************************************
// *                                                                   *
// * (c) 1997-2004 Andreas Ziermann LOET.DE                            *
// *                                                                   *
// * These files are for reference only. You may not use these files   *
// * for commercial purpose use without permit.                        *
// *                                                                   *
// * b@loet.de                                                         *
// *                                                                   *
// * 1.00 2004-07-04                                                   *
// *      first released                                               *
// *                                                                   *
// *********************************************************************

#ifndef __AZARGUMENTS_H__
#define __AZARGUMENTS_H__

enum AZArgumentKeyState
{
  F_OPTIONAL         =1, // die Angabe des Argument kann weggelassen werden
                         // wenn weggelassen wird das Argument mit 0 
                         // initialisiert
  F_REQUIRED         =2  // dieses Argument ist notwendig
};

enum AZArgumentParamState
{
  P_NOARG,
  P_OPTIONAL,
  P_REQUIRED
};

class AZStringArgument
{
  friend class AZArguments;
protected:
  typedef AZList <AZStringArgument> KeyList;
  AZArguments        & args;
  long                 param;
  AZString             paramStr;
  AZString             key,shortCut,longHelp,shortHelp;
  AZArgumentKeyState   kState;
  AZArgumentParamState pState;

  KeyList              need;
  KeyList              dontNeed;
  KeyList              atLeastOneOfThese;
  
public:
  AZStringArgument(AZArguments &_args,
             const char *keyName,const char *shortCut,
             const char *longHelp,const char *shortHelp,
             AZArgumentKeyState keyState, AZArgumentParamState paramState);
  virtual ~AZStringArgument() {}
  int operator ==(const char *_keyconst) const;

  AZArgumentParamState GetParamState() const { return pState; }
  AZArgumentKeyState   GetKeyState()   const { return kState; }
  virtual void Set(const char * _param) { param=true; paramStr=_param; }
  bool Get() const { return param?true:false; }
  const char *GetChars() const { return paramStr; }

  const AZString & GetKey() const       { return key; }
  const AZString & GetShortCut() const  { return shortCut; }
  const AZString & GetLongHelp() const  { return longHelp; }
  const AZString & GetShortHelp() const { return shortHelp; }
  
  // nun die abhaengigkeiten
  // die option braucht auch die andere option
  void Need(const AZStringArgument &);
  // die optionen schliessen sich gegenseitig aus
  // die andere option weiss das dann auch
  void DontNeed(const AZStringArgument &);
  // eine der optionen muss aber angegeben werden
  // die andere option weiss das dann auch
  void AtLeastOneOfThese(const AZStringArgument &);
};

class AZFlagArgument: public AZStringArgument
{
public:
  AZFlagArgument(AZArguments &args,const char *keyName,const char *shortCut,
                const char *lh);
  // ,const char *sh);
  // wozu ???              AZArgumentKeyState state);
  void Set(const char * _param);
};

class AZLongArgument: public AZStringArgument
{
  long paramInt;
public:
  AZLongArgument(AZArguments &_args,
             const char *keyName,const char *shortCut,
             const char *longHelp,const char *shortHelp,
             AZArgumentKeyState keyState, AZArgumentParamState paramState);
  void Set(const char *);
  long GetLong() const { return paramInt; }
};

class AZULongArgument: public AZStringArgument
{
  unsigned long paramInt;
public:
  AZULongArgument(AZArguments &_args,
             const char *keyName,const char *shortCut,
             const char *longHelp,const char *shortHelp,
             AZArgumentKeyState keyState, AZArgumentParamState paramState);
  void Set(const char *);
  long GetULong() const { return paramInt; }
};

class AZArguments : public AZStrError
{
  friend class AZStringArgument;
  typedef AZList <AZStringArgument> KeyList;

  int           argc;
  char **       argv;
  char **       env;
  AZString      errorTxt;
  AZString      desc;
  AZString      ver;
  KeyList       args;

  // default optionen, die jedes programm hat
  AZStringArgument *rc;
  AZFlagArgument   *help;
  AZFlagArgument   *quiet;
  AZULongArgument  *debug;
  AZFlagArgument   *version;

  AZStringArgument * FindMatchingKey(const char *_key);
  void InsertArgument(AZStringArgument &arg) { args.AddTail(&arg); }

  // eigentliche funktion zur bearbeitung der argumente
  void _ProcessArguments();
  // und am schluss test auf die verschiedenen abhaengigkeiten
  void TestForDepencies();

  // prints syntax (short-help)
  void ShortHelp() const;
  // print full description
  void LongHelp() const;

  AZString GetProgrammName() const;

  // vorinitialisierung der argumente aus den rc-files
  void InitFromRCFiles();
protected:
  // nach dem kontruktor!!!! diese funktion aufrufen!!!
  // hier wird noch eine nachbehandlung durchgefuehrt
  // zwecks hilfe/error...
  // im fehlerfalle wird exit(error) aufgerufen
  // und kommt also nur zurueck, wenn alles klar ging
  // aber zu erstmal wird aus den rc-files gelesen
  void ProcessArguments();

public:
  // aus dem konstruktor kommt man nur zurueck, wenn alles
  // klar ging. siehe void ProcessArguments()
  AZArguments(int argc,char *argv[],char **env,
              const char *desc,const char *ver,word exclude_flags=0);
  ~AZArguments();
};

#define FLAG_EXCLUDE_RC_ARG        1
#define FLAG_EXCLUDE_HELP_ARG      2
#define FLAG_EXCLUDE_QUIET_ARG     4
#define FLAG_EXCLUDE_DEBUG_ARG     8
#define FLAG_EXCLUDE_VERSION_ARG  16



extern const AZArguments * __azargs;

#define GIVEARGS(MyArgs,myargs) assert(__azargs); \
                               MyArgs & myargs=*(MyArgs*)__azargs

#endif

